@@ -49,6 +49,27 @@ class CacheExceptionDpkgInterrupted(CacheException):
     pass
 
 
+class PackageRemovalDeniedException(Exception):
+    def __init__(self, package):
+        self.package = package
+
+        msg = _(
+            f'The package {self.package} is marked for removal '
+            'but it is in the removal deny list.'
+        )
+
+        # If additional details can be given to the user explaining the
+        # situtation, provide them here.
+        if re.compile(r'^postgresql-[0-9]*$').match(self.package):
+            msg += _(
+                '\n\nTo prevent data loss, postgresql packages are not removed '
+                'automatically during the upgrade. If you are certain you '
+                f'no longer need {self.package}, you can manually remove it '
+                'and try the upgrade again.'
+            )
+
+        super().__init__(msg)
+
 def estimate_kernel_initrd_size_in_boot():
     """estimate the amount of space used by the kernel and initramfs in /boot,
     including a safety margin
@@ -69,11 +90,11 @@ def estimate_kernel_initrd_size_in_boot():
     if kernel == 0:
         logging.warning(
             "estimate_kernel_initrd_size_in_boot() returned '0' for kernel?")
-        kernel = 28*1024*1024
+        kernel = 16*1024*1024
     if initrd == 0:
         logging.warning(
             "estimate_kernel_initrd_size_in_boot() returned '0' for initrd?")
-        initrd = 45*1024*1024
+        initrd = 175*1024*1024
     # add small safety buffer
     kernel += 1*1024*1024
     # safety buffer as a percentage of the existing initrd's size
@@ -115,7 +136,15 @@ class MyCache(apt.Cache):
     HoldReInstReq = 3
 
     # init
-    def __init__(self, config, view, quirks, progress=None, lock=True):
+    def __init__(self,
+                 config,
+                 view,
+                 quirks,
+                 progress=None,
+                 lock=True,
+                 from_dist=None,
+                 to_dist=None):
+
         self.to_install = []
         self.to_remove = []
         self.view = view
@@ -124,6 +153,8 @@ class MyCache(apt.Cache):
         self.partialUpgrade = False
         self.config = config
         self.metapkgs = self.config.getlist("Distro", "MetaPkgs")
+        self.from_dist = from_dist
+        self.to_dist = to_dist
         # acquire lock
         self._listsLock = -1
         if lock:
@@ -139,28 +170,14 @@ class MyCache(apt.Cache):
         # Do not create the cache until we know it is not locked
         apt.Cache.__init__(self, progress)
         # a list of regexp that are not allowed to be removed
-        self.removal_blacklist = config.getListFromFile("Distro", "RemovalBlacklistFile")
+        self.removal_denylist = config.getListFromFile("Distro", "RemovalDenylistFile")
+        # the linux metapackage should not be removed
+        self.linux_metapackage = self.quirks._get_linux_metapackage(self, False)
         self.uname = Popen(["uname", "-r"], stdout=PIPE,
                            universal_newlines=True).communicate()[0].strip()
         self._initAptLog()
-        # from hardy on we use recommends by default, so for the
-        # transition to the new dist we need to enable them now
-        if (config.get("Sources", "From") == "hardy" and
-            not "RELEASE_UPGRADE_NO_RECOMMENDS" in os.environ):
-            apt_pkg.config.set("APT::Install-Recommends", "true")
-
-
         apt_pkg.config.set("APT::AutoRemove::SuggestsImportant", "false")
 
-    def _apply_dselect_upgrade(self):
-        """ honor the dselect install state """
-        for pkg in self:
-            if pkg.is_installed:
-                continue
-            if pkg._pkg.selected_state == apt_pkg.SELSTATE_INSTALL:
-                # upgrade() will take care of this
-                pkg.mark_install(auto_inst=False, auto_fix=False)
-
     @property
     def req_reinstall_pkgs(self):
         " return the packages not downloadable packages in reqreinst state "
@@ -256,6 +273,20 @@ class MyCache(apt.Cache):
         """ get the size of the additional required space on the fs """
         return self._depcache.usr_size
     @property
+    def additional_required_space_for_snaps(self):
+        """ get the extra size needed to install the snap replacements """
+        try:
+            # update-manager uses DistUpgradeCache.MyCache as the base class
+            # of its own MyCache version - but without actually calling our
+            # constructor at all. This causes that the MyCache version from
+            # update-manager has no self.quirks attribute while still calling
+            # our default version of checkFreeSpace(). Since extra_snap_space
+            # is only used on dist-upgrades, let's just not care and return 0
+            # in this weird, undocumented case.
+            return self.quirks.extra_snap_space
+        except AttributeError:
+            return 0
+    @property
     def is_broken(self):
         """ is the cache broken """
         return self._depcache.broken_count > 0
@@ -292,7 +323,7 @@ class MyCache(apt.Cache):
             self.release_lock()
         apt.Cache.commit(self, fprogress, iprogress)
 
-    def release_lock(self, pkgSystemOnly=True):
+    def release_lock(self):
         if self.lock:
             try:
                 apt_pkg.pkgsystem_unlock()
@@ -300,7 +331,7 @@ class MyCache(apt.Cache):
             except SystemError as e:
                 logging.debug("failed to SystemUnLock() (%s) " % e)
 
-    def get_lock(self, pkgSystemOnly=True):
+    def get_lock(self):
         if not self.lock:
             try:
                 apt_pkg.pkgsystem_lock()
@@ -319,16 +350,28 @@ class MyCache(apt.Cache):
             return False
         return ver.downloadable
 
-    def pkg_auto_removable(self, pkg):
-        """ check if the pkg is auto-removable """
-        return (pkg.is_installed and
-                self._depcache.is_garbage(pkg._pkg))
-
     def fix_broken(self):
         """ try to fix broken dependencies on the system, may throw
             SystemError when it can't"""
         return self._depcache.fix_broken()
 
+    def get_annotated_changes(self):
+        """Get a set of (name, change) string pairs"""
+        changes = set()
+        for pkg in self.get_changes():
+            if pkg.marked_install:
+                change = "install"
+            elif pkg.marked_upgrade:
+                change = "upgrade"
+            elif pkg.marked_delete:
+                change = "delete"
+            else:
+                continue
+
+            changes.add((pkg.name, change))
+
+        return changes
+
     def create_snapshot(self):
         """ create a snapshot of the current changes """
         self.to_install = []
@@ -386,7 +429,7 @@ class MyCache(apt.Cache):
         logging.debug("need_server_mode(): can not find a desktop meta package or key deps, running in server mode")
         return True
 
-    def sanity_check(self, view):
+    def coherence_check(self, view):
         """ check if the cache is ok and if the required metapkgs
             are installed
         """
@@ -404,10 +447,10 @@ class MyCache(apt.Cache):
                 return False
         return True
 
-    def mark_install(self, pkg, reason=""):
+    def mark_install(self, pkg, reason="", **flags):
         logging.debug("Installing '%s' (%s)" % (pkg, reason))
         if pkg in self:
-            self[pkg].mark_install()
+            self[pkg].mark_install(**flags)
             if not (self[pkg].marked_install or self[pkg].marked_upgrade):
                 logging.error("Installing/upgrading '%s' failed" % pkg)
                 #raise SystemError("Installing '%s' failed" % pkg)
@@ -423,10 +466,10 @@ class MyCache(apt.Cache):
                 return False
         return True
 
-    def mark_remove(self, pkg, reason=""):
+    def mark_remove(self, pkg, reason="", **flags):
         logging.debug("Removing '%s' (%s)" % (pkg, reason))
         if pkg in self:
-            self[pkg].mark_delete()
+            self[pkg].mark_delete(**flags)
 
     def mark_purge(self, pkg, reason=""):
         logging.debug("Purging '%s' (%s)" % (pkg, reason))
@@ -452,27 +495,31 @@ class MyCache(apt.Cache):
                 for pkgname in self.config.getlist(key, "KeepInstalledPkgs"):
                     self._keep_installed(pkgname, "%s KeepInstalledPkgs rule" % key)
 
-        # only enforce section if we have a network. Otherwise we run
-        # into CD upgrade issues for installed language packs etc
-        if self.config.get("Options", "withNetwork") == "True":
-            logging.debug("Running KeepInstalledSection rules")
-            # now the KeepInstalledSection code
-            for section in self.config.getlist("Distro", "KeepInstalledSection"):
-                for pkg in self:
-                    if (pkg.candidate and pkg.candidate.downloadable
-                        and pkg.marked_delete and pkg.section == section):
-                        self._keep_installed(pkg.name, "Distro KeepInstalledSection rule: %s" % section)
-            for key in self.metapkgs:
-                if key in self and (self[key].is_installed or
-                                    self[key].marked_install):
-                    for section in self.config.getlist(key, "KeepInstalledSection"):
-                        for pkg in self:
-                            if (pkg.candidate and pkg.candidate.downloadable
-                                and pkg.marked_delete and
-                                pkg.section == section):
-                                self._keep_installed(pkg.name, "%s KeepInstalledSection rule: %s" % (key, section))
+        logging.debug("Running KeepInstalledSection rules")
+        # now the KeepInstalledSection code
+        for section in self.config.getlist("Distro", "KeepInstalledSection"):
+            for pkg in self:
+                if (pkg.candidate and pkg.candidate.downloadable
+                    and pkg.marked_delete
+                    and pkg.candidate.section == section):
+                    self._keep_installed(pkg.name, "Distro KeepInstalledSection rule: %s" % section)
+        for key in self.metapkgs:
+            if key in self and (self[key].is_installed or
+                                self[key].marked_install):
+                for section in self.config.getlist(key, "KeepInstalledSection"):
+                    for pkg in self:
+                        if (pkg.candidate and pkg.candidate.downloadable
+                            and pkg.marked_delete and
+                            pkg.candidate.section == section):
+                            self._keep_installed(pkg.name, "%s KeepInstalledSection rule: %s" % (key, section))
 
 
+    def pre_upgrade_rule(self):
+        " run before the upgrade was done in the cache "
+        # run the quirks handlers
+        if not self.partialUpgrade:
+            self.quirks.PreDistUpgradeCache()
+
     def post_upgrade_rule(self):
         " run after the upgrade was done in the cache "
         for (rule, action) in [("Install", self.mark_install),
@@ -489,38 +536,7 @@ class MyCache(apt.Cache):
                         action(pkg, "%s PostUpgrade%s rule" % (key, rule))
         # run the quirks handlers
         if not self.partialUpgrade:
-            self.quirks.run("PostDistUpgradeCache")
-
-    def identifyObsoleteKernels(self):
-        # we have a funny policy that we remove security updates
-        # for the kernel from the archive again when a new ABI
-        # version hits the archive. this means that we have
-        # e.g.
-        # linux-image-2.6.24-15-generic
-        # is obsolete when
-        # linux-image-2.6.24-19-generic
-        # is available
-        # ...
-        # This code tries to identify the kernels that can be removed
-        logging.debug("identifyObsoleteKernels()")
-        obsolete_kernels = set()
-        version = self.config.get("KernelRemoval", "Version")
-        basenames = self.config.getlist("KernelRemoval", "BaseNames")
-        types = self.config.getlist("KernelRemoval", "Types")
-        for pkg in self:
-            for base in basenames:
-                basename = "%s-%s-" % (base, version)
-                for type in types:
-                    if (pkg.name.startswith(basename) and
-                        pkg.name.endswith(type) and
-                        pkg.is_installed):
-                        if (pkg.name == "%s-%s" % (base, self.uname)):
-                            logging.debug("skipping running kernel %s" % pkg.name)
-                            continue
-                        logging.debug("removing obsolete kernel '%s'" % pkg.name)
-                        obsolete_kernels.add(pkg.name)
-        logging.debug("identifyObsoleteKernels found '%s'" % obsolete_kernels)
-        return obsolete_kernels
+            self.quirks.PostDistUpgradeCache()
 
     def checkForNvidia(self):
         """
@@ -563,21 +579,13 @@ class MyCache(apt.Cache):
             logging.error("NvidiaDetection returned a error: %s" % e)
         return False
 
-
-    def _has_kernel_headers_installed(self):
-        for pkg in self:
-            if (pkg.name.startswith("linux-headers-") and
-                pkg.is_installed):
-                return True
-        return False
-
     def checkForKernel(self):
         """ check for the running kernel and try to ensure that we have
             an updated version
         """
         logging.debug("Kernel uname: '%s' " % self.uname)
         try:
-            (version, build, flavour) = self.uname.split("-")
+            (version, _, _) = self.uname.split("-")
         except Exception as e:
             logging.warning("Can't parse kernel uname: '%s' (self compiled?)" % e)
             return False
@@ -627,8 +635,14 @@ class MyCache(apt.Cache):
         t = threading.Thread(target=self.updateGUI, args=(self.view, lock,))
         t.start()
         try:
-            # mvo: disabled as it casues to many errornous installs
-            #self._apply_dselect_upgrade()
+            # run PreDistUpgradeCache quirks
+            self.pre_upgrade_rule()
+
+            # install missing meta-packages (if not in server upgrade mode)
+            self._keepBaseMetaPkgsInstalled(view)
+            if not serverMode:
+                # if this fails, a system error is raised
+                self._installMetaPkgs(view)
 
             # upgrade (and make sure this way that the cache is ok)
             self.upgrade(True)
@@ -652,45 +666,39 @@ class MyCache(apt.Cache):
             # and if we have some special rules
             self.post_upgrade_rule()
 
-            # install missing meta-packages (if not in server upgrade mode)
-            self._keepBaseMetaPkgsInstalled(view)
-            if not serverMode:
-                # if this fails, a system error is raised
-                self._installMetaPkgs(view)
-
             # see if it all makes sense, if not this function raises
             self._verifyChanges()
 
-        except SystemError as e:
-            # this should go into a finally: line, see below for the
-            # rationale why it doesn't
-            lock.release()
-            t.join()
-            # the most likely problem is the 3rd party pkgs so don't address
-            # foreignPkgs and devRelease being True
+            if self.is_broken:
+                raise SystemError(_("Broken packages after upgrade: %s") % ", ".join(p.name for p in self if p.is_inst_broken or p.is_now_broken))
+
+        except (SystemError, PackageRemovalDeniedException) as e:
             details =  _("An unresolvable problem occurred while "
                          "calculating the upgrade.\n\n ")
-            if self.config.get("Options", "foreignPkgs") == "True":
-                details += _("This was likely caused by:\n"
-                             " * Unofficial software packages not provided by Ubuntu\n"
-                             "Please use the tool 'ppa-purge' from the ppa-purge \n"
-                             "package to remove software from a Launchpad PPA and \n"
-                             "try the upgrade again.\n"
-                             "\n")
-            elif self.config.get("Options", "foreignPkgs") == "False" and \
-                self.config.get("Options", "devRelease") == "True":
-                details +=  _("This was caused by:\n"
-                              " * Upgrading to a pre-release version of Ubuntu\n"
-                              "This is most likely a transient problem, \n"
-                              "please try again later.\n")
+
+            foreign_packages = self.foreign_packages()
+
             # we never have partialUpgrades (including removes) on a stable system
             # with only ubuntu sources so we do not recommend reporting a bug
             if partialUpgrade:
                 details += _("This is most likely a transient problem, "
                              "please try again later.")
+
+            elif isinstance(e, PackageRemovalDeniedException):
+                details += str(e) + '\n\n'
+
+            elif foreign_packages:
+                details += _(
+                    "This may be caused by unofficial packages installed on "
+                    "the system. Please try replacing the following packages "
+                    "with official versions from the Ubuntu archive, and then "
+                    "try the upgrade again.\n\n"
+                )
+                details += '\n'.join(foreign_packages)
+
             else:
                 details += _("If none of this applies, then please report this bug using "
-                             "the command 'ubuntu-bug ubuntu-release-upgrader-core' in a terminal.")
+                             "the command 'ubuntu-bug ubuntu-release-upgrader-core' in a terminal. ")
                 details += _("If you want to investigate this yourself the log files in "
                              "'/var/log/dist-upgrade' will contain details about the upgrade. "
                              "Specifically, look at 'main.log' and 'apt.log'.")
@@ -705,12 +713,10 @@ class MyCache(apt.Cache):
             # the withResolverLog decorator
             self._startAptResolverLog()
             return False
-        # would be nice to be able to use finally: here, but we need
-        # to run on python2.4 too
-        #finally:
-        # wait for the gui-update thread to exit
-        lock.release()
-        t.join()
+        finally:
+            # wait for the gui-update thread to exit
+            lock.release()
+            t.join()
 
         # check the trust of the packages that are going to change
         untrusted = []
@@ -744,7 +750,7 @@ class MyCache(apt.Cache):
             if b:
                 logging.warning("AllowUnauthenticated set!")
                 return True
-        except configparser.NoOptionError as e:
+        except configparser.NoOptionError:
             pass
         if len(downgrade) > 0:
             downgrade.sort()
@@ -770,22 +776,22 @@ class MyCache(apt.Cache):
 
     def _verifyChanges(self):
         """ this function tests if the current changes don't violate
-            our constrains (blacklisted removals etc)
+            our constraints (deny listed removals etc)
         """
         main_arch = apt_pkg.config.find("APT::Architecture")
         removeEssentialOk = self.config.getlist("Distro", "RemoveEssentialOk")
         # check changes
         for pkg in self.get_changes():
-            if pkg.marked_delete and self._inRemovalBlacklist(pkg.name):
-                logging.debug("The package '%s' is marked for removal but it's in the removal blacklist", pkg.name)
-                raise SystemError(_("The package '%s' is marked for removal but it is in the removal blacklist.") % pkg.name)
+            if pkg.marked_delete and self._inRemovalDenylist(pkg.name):
+                logging.debug("The package '%s' is marked for removal but it's in the removal deny list", pkg.name)
+                raise PackageRemovalDeniedException(pkg.name)
             if pkg.marked_delete and (
                     pkg._pkg.essential == True and
                     pkg.installed.architecture in (main_arch, "all") and
                     not pkg.name in removeEssentialOk):
                 logging.debug("The package '%s' is marked for removal but it's an ESSENTIAL package", pkg.name)
                 raise SystemError(_("The essential package '%s' is marked for removal.") % pkg.name)
-        # check bad-versions blacklist
+        # check bad-versions deny list
         badVersions = self.config.getlist("Distro", "BadVersions")
         for bv in badVersions:
             (pkgname, ver) = bv.split("_")
@@ -793,7 +799,7 @@ class MyCache(apt.Cache):
                 self[pkgname].candidate.version == ver and
                 (self[pkgname].marked_install or
                  self[pkgname].marked_upgrade)):
-                raise SystemError(_("Trying to install blacklisted version '%s'") % bv)
+                raise SystemError(_("Trying to install deny listed version '%s'") % bv)
         return True
 
     def _lookupPkgRecord(self, pkg):
@@ -831,6 +837,9 @@ class MyCache(apt.Cache):
                         tasks[task].add(pkg.name)
         for task in tasks:
             installed = True
+            ignored_tasks = self.config.getlist("Distro", "IgnoredTasks")
+            if task in ignored_tasks:
+                installed = False
             for pkgname in tasks[task]:
                 if not self[pkgname].is_installed:
                     installed = False
@@ -883,6 +892,7 @@ class MyCache(apt.Cache):
         # we never go without ubuntu-base
         for pkg in self.config.getlist("Distro", "BaseMetaPkgs"):
             self[pkg].mark_install()
+            apt.ProblemResolver(self).protect(self[pkg])
 
         # every meta-pkg that is installed currently, will be marked
         # install (that result in a upgrade and removes a mark_delete)
@@ -893,6 +903,7 @@ class MyCache(apt.Cache):
                     self[key].is_upgradable):
                     logging.debug("Marking '%s' for upgrade" % key)
                     self[key].mark_upgrade()
+                    apt.ProblemResolver(self).protect(self[key])
             except SystemError as e:
                 # warn here, but don't fail, its possible that meta-packages
                 # conflict (like ubuntu-desktop vs xubuntu-desktop) LP: #775411
@@ -936,61 +947,74 @@ class MyCache(apt.Cache):
             return False
         return True
 
-    def _inRemovalBlacklist(self, pkgname):
-        for expr in self.removal_blacklist:
+    def _inRemovalDenylist(self, pkgname):
+        for expr in self.removal_denylist:
             if re.compile(expr).match(pkgname):
-                logging.debug("blacklist expr '%s' matches '%s'" % (expr, pkgname))
+                logging.debug("denylist expr '%s' matches '%s'" %
+                              (expr, pkgname))
                 return True
         return False
 
-    @withResolverLog
-    def tryMarkObsoleteForRemoval(self, pkgname, remove_candidates, foreign_pkgs):
-        #logging.debug("tryMarkObsoleteForRemoval(): %s" % pkgname)
-        # sanity check, first see if it looks like a running kernel pkg
+    def isRemoveCandidate(self, pkgname, foreign_pkgs):
+        # coherence check, first see if it looks like a running kernel pkg
+        if pkgname in foreign_pkgs:
+            logging.debug("skipping foreign pkg '%s'" % pkgname)
+            return False
         if pkgname.endswith(self.uname):
             logging.debug("skipping running kernel pkg '%s'" % pkgname)
             return False
-        if self._inRemovalBlacklist(pkgname):
-            logging.debug("skipping '%s' (in removalBlacklist)" % pkgname)
+        if pkgname == self.linux_metapackage:
+            logging.debug("skipping kernel metapackage '%s'" % pkgname)
+            return False
+        if self._inRemovalDenylist(pkgname):
+            logging.debug("skipping '%s' (in removalDenylist)" % pkgname)
             return False
         # ensure we honor KeepInstalledSection here as well
         for section in self.config.getlist("Distro", "KeepInstalledSection"):
-            if pkgname in self and self[pkgname].section == section:
+            if (pkgname in self and self[pkgname].installed and
+                    self[pkgname].installed.section == section):
                 logging.debug("skipping '%s' (in KeepInstalledSection)" % pkgname)
                 return False
+
+        return True
+
+    @withResolverLog
+    def tryMarkObsoleteForRemoval(self, pkgname, remove_candidates, forced_obsoletes, auto_fix):
+        #logging.debug("tryMarkObsoleteForRemoval(): %s" % pkgname)
         # if we don't have the package anyway, we are fine (this can
         # happen when forced_obsoletes are specified in the config file)
+        if pkgname not in remove_candidates:
+            return False
         if pkgname not in self:
             #logging.debug("package '%s' not in cache" % pkgname)
             return True
         # check if we want to purge
         try:
             purge = self.config.getboolean("Distro", "PurgeObsoletes")
-        except configparser.NoOptionError as e:
+        except configparser.NoOptionError:
             purge = False
 
-        # this is a delete candidate, only actually delete,
-        # if it dosn't remove other packages depending on it
+        # if this package has not been forced obsolete, only
+        # delete it if it doesn't remove other dependents
         # that are not obsolete as well
-        actiongroup = apt_pkg.ActionGroup(self._depcache)
-        # just make pyflakes shut up, later we should use
-        # with self.actiongroup():
-        actiongroup
-        self.create_snapshot()
+        if auto_fix:
+            self.create_snapshot()
         try:
-            self[pkgname].mark_delete(purge=purge)
+            self[pkgname].mark_delete(purge=purge, auto_fix=auto_fix)
             self.view.processEvents()
-            #logging.debug("marking '%s' for removal" % pkgname)
-            for pkg in self.get_changes():
-                if (pkg.name not in remove_candidates or
-                      pkg.name in foreign_pkgs or
-                      self._inRemovalBlacklist(pkg.name)):
-                    logging.debug("package '%s' has unwanted removals, skipping" % pkgname)
-                    self.restore_snapshot()
-                    return False
+            if auto_fix:
+                if pkgname in forced_obsoletes:
+                    return True
+                #logging.debug("marking '%s' for removal" % pkgname)
+                for pkg in self.get_changes():
+                    if pkg.name not in remove_candidates:
+                        logging.debug("package '%s' produces an unwanted removal '%s', skipping" % (pkgname, pkg.name))
+                        self.restore_snapshot()
+                        return False
         except (SystemError, KeyError) as e:
             logging.warning("_tryMarkObsoleteForRemoval failed for '%s' (%s: %s)" % (pkgname, repr(e), e))
-            self.restore_snapshot()
+            if auto_fix:
+                self.restore_snapshot()
             return False
         return True
 
@@ -1022,53 +1046,45 @@ class MyCache(apt.Cache):
                 unused_dependencies.add(pkg.name)
         return unused_dependencies
 
-    def get_installed_demoted_packages(self):
-        """ return list of installed and demoted packages
+    def foreign_packages(self):
+        '''
+        Check for packages that do not have 'Ubuntu' as their origin.
+        Return this as a dict with <package name> -> <package file>.
 
-            If a demoted package is a automatic install it will be skipped
-        """
-        demotions = set()
-        demotions_file = self.config.get("Distro", "Demotions")
-        if os.path.exists(demotions_file):
-            with open(demotions_file) as demotions_f:
-                for line in demotions_f:
-                    if not line.startswith("#"):
-                        demotions.add(line.strip())
-        installed_demotions = set()
-        for demoted_pkgname in demotions:
-            if demoted_pkgname not in self:
-                continue
-            pkg = self[demoted_pkgname]
-            if (not pkg.is_installed or
-                self._depcache.is_auto_installed(pkg._pkg) or
-                pkg.marked_delete):
-                continue
-            installed_demotions.add(pkg)
-        return list(installed_demotions)
+        If an installed package has a candidate that is not downloadable,
+        but *some* version of the package is downloadable, consider that
+        package foreign. In such cases, the package will have a value of None
+        in the returned dict.
+        '''
+        foreign = dict()
 
-    def _getForeignPkgs(self, allowed_origin, fromDist, toDist):
-        """ get all packages that are installed from a foreign repo
-            (and are actually downloadable)
-        """
-        foreign_pkgs = set()
         for pkg in self:
-            if pkg.is_installed and self.downloadable(pkg):
-                if not pkg.candidate:
-                    continue
-                # assume it is foreign and see if it is from the
-                # official archive
-                foreign = True
-                for origin in pkg.candidate.origins:
-                    # FIXME: use some better metric here
-                    if fromDist in origin.archive and \
-                           origin.origin == allowed_origin:
-                        foreign = False
-                    if toDist in origin.archive and \
-                           origin.origin == allowed_origin:
-                        foreign = False
-                if foreign:
-                    foreign_pkgs.add(pkg.name)
-        return foreign_pkgs
+            if not (pkg.is_installed and pkg.candidate):
+                continue
+
+            if self.downloadable(pkg):
+                for (pkg_file, _) in pkg.candidate._cand.file_list:
+                    # Only consider configured apt sources
+                    if pkg_file.not_source:
+                        continue
+
+                    if not (
+                        pkg_file.origin == 'Ubuntu' and
+                        (
+                            pkg_file.archive.startswith(self.from_dist) or
+                            pkg_file.archive.startswith(self.to_dist)
+                        )
+                    ):
+                        # This is a foreign package
+                        foreign[pkg.name] = pkg_file
+                        break
+
+            elif self.anyVersionDownloadable(pkg):
+                # This can happen, for example, if a package is installed from
+                # a PPA, and that version is newer than what is in the archive.
+                foreign[pkg.name] = None
+
+        return foreign
 
     def checkFreeSpace(self, snapshots_in_use=False):
         """
@@ -1106,7 +1122,7 @@ class MyCache(apt.Cache):
         with open("/proc/mounts") as mounts:
             for line in mounts:
                 try:
-                    (what, where, fs, options, a, b) = line.split()
+                    (_, where, _, options, a, b) = line.split()
                 except ValueError as e:
                     logging.debug("line '%s' in /proc/mounts not understood (%s)" % (line, e))
                     continue
@@ -1115,14 +1131,7 @@ class MyCache(apt.Cache):
         # make sure mounted is sorted by longest path
         mounted.sort(key=len, reverse=True)
         archivedir = apt_pkg.config.find_dir("Dir::Cache::archives")
-        aufs_rw_dir = "/tmp/"
-        if (hasattr(self, "config") and
-            self.config.getWithDefault("Aufs", "Enabled", False)):
-            aufs_rw_dir = self.config.get("Aufs", "RWDir")
-            if not os.path.exists(aufs_rw_dir):
-                os.makedirs(aufs_rw_dir)
-        logging.debug("cache aufs_rw_dir: %s" % aufs_rw_dir)
-        for d in ["/", "/usr", "/var", "/boot", archivedir, aufs_rw_dir, "/home", "/tmp/"]:
+        for d in ["/", "/usr", "/var", "/boot", archivedir, "/home", "/tmp/"]:
             d = os.path.realpath(d)
             fs_id = make_fs_id(d)
             if os.path.exists(d):
@@ -1165,16 +1174,6 @@ class MyCache(apt.Cache):
         # /usr is assumed to get *all* of the install space (incorrect,
         #      but as good as we can do currently + safety buffer
         # /     has a small safety buffer as well
-        required_for_aufs = 0.0
-        if (hasattr(self, "config") and
-            self.config.getWithDefault("Aufs", "Enabled", False)):
-            logging.debug("taking aufs overlay into space calculation")
-            aufs_rw_dir = self.config.get("Aufs", "RWDir")
-            # if we use the aufs rw overlay all the space is consumed
-            # the overlay dir
-            for pkg in self:
-                if pkg.marked_upgrade or pkg.marked_install:
-                    required_for_aufs += pkg.candidate.installed_size
 
         # add old size of the package if we use snapshots
         required_for_snapshots = 0.0
@@ -1188,12 +1187,13 @@ class MyCache(apt.Cache):
         # sum up space requirements
         for (dir, size) in [(archivedir, self.required_download),
                             ("/usr", self.additional_required_space),
+                            # this is only >0 for the deb-to-snap quirks
+                            ("/var", self.additional_required_space_for_snaps),
                             # plus 50M safety buffer in /usr
                             ("/usr", 50*1024*1024),
                             ("/boot", space_in_boot),
                             ("/tmp", 5*1024*1024),   # /tmp for dkms LP: #427035
                             ("/", 10*1024*1024),     # small safety buffer /
-                            (aufs_rw_dir, required_for_aufs),
                             # if snapshots are in use
                             ("/usr", required_for_snapshots),
                            ]:
@@ -1227,22 +1227,3 @@ class MyCache(apt.Cache):
             logging.error("Not enough free space: %s" % [str(i) for i in required_list])
             raise NotEnoughFreeSpaceError(list(required_list.values()))
         return True
-
-
-if __name__ == "__main__":
-    import sys
-    from .DistUpgradeConfigParser import DistUpgradeConfig
-    from .DistUpgradeView import DistUpgradeView
-    print("foo")
-    c = MyCache(DistUpgradeConfig("."), DistUpgradeView(), None)
-    #c.checkForNvidia()
-    #print(c._identifyObsoleteKernels())
-    print(c.checkFreeSpace())
-    sys.exit()
-
-    c.clear()
-    c.create_snapshot()
-    c.installedTasks
-    c.installTasks(["ubuntu-desktop"])
-    print(c.get_changes())
-    c.restore_snapshot()
