@@ -11,11 +11,12 @@ from optparse import OptionParser
 import re
 import gettext
 import distro_info
-import subprocess
+from contextlib import redirect_stderr
 
 
 SYNAPTIC_PINFILE = "/var/lib/synaptic/preferences"
 OS_RELEASE_PATH = "/etc/os-release"
+PRO_ESM_CACHE_DIR = "/var/lib/ubuntu-advantage/apt-esm/"
 
 
 def _get_info_from_os_release(key):
@@ -42,22 +43,9 @@ def _get_info_from_os_release(key):
         )
 
 
-def _get_output_from_lsb_release(lsb_option):
-    " get info from lsb_release output "
-    return subprocess.check_output(
-        ["lsb_release", lsb_option, "-s"], universal_newlines=True
-    ).strip()
-
-
 def get_distro():
     " get distro name "
-    try:
-        return _get_info_from_os_release(key="UBUNTU_CODENAME")
-    except Exception:
-        # If the system does not have os-release file or does not have the
-        # required entry in it, we will get the distro name from lsb_release
-        # command
-        return _get_output_from_lsb_release("-c")
+    return _get_info_from_os_release(key="UBUNTU_CODENAME")
 
 
 DISTRO = get_distro()
@@ -78,13 +66,7 @@ def _handleException(type, value, tb):
 
 def get_distro_version():
     " get distro version "
-    try:
-        return _get_info_from_os_release(key="VERSION_ID").replace('"', "")
-    except Exception:
-        # If the system does not have os-release file or does not have the
-        # required entry in it, we will get the distro name from lsb_release
-        # command
-        return _get_output_from_lsb_release("-r")
+    return _get_info_from_os_release(key="VERSION_ID").replace('"', "")
 
 
 def clean(cache, depcache):
@@ -241,21 +223,23 @@ def write_human_readable_summary(outstream, upgrades, security_updates,
                                  esm_infra_updates, esm_apps_updates,
                                  have_esm_infra, have_esm_apps,
                                  disabled_esm_infra_updates,
-                                 disabled_esm_apps_updates):
+                                 disabled_esm_apps_updates,
+                                 hide_esm_messages=False):
 
-    " write out human summary to outstream "
+    " write out human summary summary to outstream "
     esm_distro = is_esm_distro()
     lts_distro = is_lts_distro()
 
-    if have_esm_infra is not None and esm_distro:
-        _output_esm_service_status(
-            outstream, have_esm_infra, service_type="Infrastructure"
-        )
+    if not hide_esm_messages:
+        if have_esm_infra is not None and esm_distro:
+            _output_esm_service_status(
+                outstream, have_esm_infra, service_type="Infrastructure"
+            )
 
-    if have_esm_apps is not None and lts_distro and not esm_distro:
-        _output_esm_service_status(
-            outstream, have_esm_apps, service_type="Applications"
-        )
+        if have_esm_apps is not None and lts_distro and not esm_distro:
+            _output_esm_service_status(
+                outstream, have_esm_apps, service_type="Applications"
+            )
 
     outstream.write(
         gettext.dngettext("update-notifier",
@@ -264,10 +248,11 @@ def write_human_readable_summary(outstream, upgrades, security_updates,
                           upgrades) % upgrades
     )
 
-    _output_esm_package_count(
-        outstream, service_type="Infra", esm_pkg_count=esm_infra_updates)
-    _output_esm_package_count(
-        outstream, service_type="Apps", esm_pkg_count=esm_apps_updates)
+    if not hide_esm_messages:
+        _output_esm_package_count(
+            outstream, service_type="Infra", esm_pkg_count=esm_infra_updates)
+        _output_esm_package_count(
+            outstream, service_type="Apps", esm_pkg_count=esm_apps_updates)
 
     if security_updates > 0:
         outstream.write("\n")
@@ -290,14 +275,22 @@ def write_human_readable_summary(outstream, upgrades, security_updates,
             have_esm_apps is not None,
             not have_esm_apps,
             lts_distro,
-            not esm_distro
+            not esm_distro,
+            not hide_esm_messages,
         ]
     ):
         _output_esm_package_alert(
             outstream, service_type="Apps",
             disabled_pkg_count=disabled_esm_apps_updates)
 
-    if have_esm_infra is not None and not have_esm_infra and esm_distro:
+    if all(
+        [
+            have_esm_infra is not None,
+            not have_esm_infra,
+            esm_distro,
+            not hide_esm_messages,
+        ]
+    ):
         _output_esm_package_alert(
             outstream, service_type="Infra",
             disabled_pkg_count=disabled_esm_infra_updates,
@@ -345,6 +338,30 @@ def has_esm_service(cache, esm_origin):
     return False
 
 
+def get_apt_pkg_esm_cache():
+    """Get an apt_pkg cache with the ubuntu-advantage-tools esm data.
+
+    Set the configuration to get the u-a-t cache, then set it back to an
+    empty configuration state and init again so other calls to Cache work as
+    expected.
+    """
+    for key in apt_pkg.config.keys():
+        if re.match(r"^Acquire", key) is None:
+            apt_pkg.config.clear(key)
+    apt_pkg.config.set("Dir", PRO_ESM_CACHE_DIR)
+    apt_pkg.init()
+    try:
+        esm_cache = apt_pkg.Cache(progress=None)
+    except apt_pkg.Error:
+        esm_cache = None
+
+    for key in apt_pkg.config.keys():
+        apt_pkg.config.clear(key)
+    apt_pkg.init()
+
+    return esm_cache
+
+
 def init():
     " init the system, be nice "
     # FIXME: do a ionice here too?
@@ -385,12 +402,8 @@ def run(options=None):
         sys.stderr.write("E: " + _("Error: Marking the upgrade (%s)") % e)
         sys.exit(-1)
 
-    try:
-        apt_pkg.config.set("Dir", "/var/lib/ubuntu-advantage/apt-esm/")
-        apt_pkg.init_system()
-        esm_cache = apt_pkg.Cache(progress=None)
-    except apt_pkg.Error:
-        esm_cache = None
+    with redirect_stderr(None):
+        esm_cache = get_apt_pkg_esm_cache()
 
     # Check if we have ESM enabled or disabled; and if it exists in the
     # first place.
@@ -473,14 +486,15 @@ def run(options=None):
                                      esm_infra_updates, esm_apps_updates,
                                      have_esm_infra, have_esm_apps,
                                      disabled_esm_infra_updates,
-                                     disabled_esm_apps_updates)
+                                     disabled_esm_apps_updates,
+                                     options.hide_esm_messages)
     else:
         # print the number of regular upgrades and the number of
         # security upgrades
         sys.stderr.write("%s;%s" % (upgrades, security_updates))
 
     # return the number of upgrades (if its used as a module)
-    return(upgrades, security_updates)
+    return (upgrades, security_updates)
 
 
 if __name__ == "__main__":
@@ -507,6 +521,12 @@ if __name__ == "__main__":
                       action="store_true",
                       dest="readable_output",
                       help=_("Show human readable output on stdout"))
+    parser.add_option("",
+                      "--no-esm-messages",
+                      action="store_true",
+                      dest="hide_esm_messages",
+                      help=_("Do not show esm related messages in human "
+                             "readable output"))
     parser.add_option("",
                       "--security-updates-unattended",
                       action="store_true",
